#import "MyEditor.h"
#import "CScintillaController.h"
#import "TXTScintillaController.h"
#import "MKScintillaController.h"
#import "SQLScintillaController.h"
#import "SciSettersGetters.h"

#import "Scintilla.h"

using namespace Scintilla;

/* offset numbers to set the scintilla editor
 * in the correct spot on the NSView
 * values obtained via experimentation
 */
#define SCINTILLA_EDITOR_X_OFFSET       26
#define SCINTILLA_EDITOR_Y_OFFSET        0
#define SCINTILLA_EDITOR_WIDTH_OFFSET  -13
#define SCINTILLA_EDITOR_HEIGHT_OFFSET   0

static void notification(intptr_t windowid, unsigned int iMessage, uintptr_t wParam, uintptr_t lParam)
{
  ScintillaView* editor;
  NSDocument* master;
  switch( iMessage )
  {
    case WM_NOTIFY:
    {
      // Parent notification. Details are passed as SCNotification structure.
      SCNotification* scn = reinterpret_cast<SCNotification*>(lParam);
      editor = reinterpret_cast<InnerView*>(scn->nmhdr.idFrom).owner;
      switch( scn->nmhdr.code )
      {
        case SCN_CHARADDED:
        case SCN_KEY:
        {
          break;
        }
        case SCN_MARGINCLICK:
        {
          int line_num = [editor getGeneralProperty: SCI_LINEFROMPOSITION parameter: scn->position];
          [editor setGeneralProperty:SCI_TOGGLEFOLD parameter: line_num value: 0];
          break;
        }
        case SCN_PAINTED:
        {
          break;
        }
        default:
        {
          break;
        }
      }

      break;
    }
    case WM_COMMAND:
    {
      ScintillaCocoa* backend = reinterpret_cast<ScintillaCocoa*>(lParam);
      editor = backend->TopContainer();
      switch (wParam >> 16)
      {
        case SCEN_SETFOCUS:
        {
          master = [editor owner];
          [master hideFind: nil];
          break;
        }
      }
      break;
    }
  }
}

@implementation MyEditor

- (id)init
{
    self = [super init];
    if( self )
    {
      _fType = FILE_TYPE_UNINIT;
      _isWrap = false;
    }
    return self;
}

/* when we quit/close with a "dirty" editor, we will be promted to save */
-(BOOL) isDocumentEdited
{
  int rc = [_sciEditor getGeneralProperty: SCI_GETMODIFY];
  if( rc )
    return YES;
  else
    return NO;
}

/* when loaded from the NIB file, lets setup the editor as a C++ editor
 * use the CScintillaController to set us up
 */
- (void)awakeFromNib
{
  ScintillaController* ed = NULL;
  NSRect newFrame = _sciHost.frame;

  /* set the scintilla editor to the right size inside the host NSView */
  newFrame.origin.x    += SCINTILLA_EDITOR_X_OFFSET;
  newFrame.origin.y    += SCINTILLA_EDITOR_Y_OFFSET;
  newFrame.size.width  += SCINTILLA_EDITOR_WIDTH_OFFSET;
  newFrame.size.height += SCINTILLA_EDITOR_HEIGHT_OFFSET;
  
  _sciEditor = [[[ScintillaView alloc] initWithFrame: newFrame] autorelease];
  
  [_sciHost addSubview: _sciEditor positioned: NSWindowBelow relativeTo: nil];
  [_sciEditor setAutoresizesSubviews: YES];
  [_sciEditor setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];

  Scintilla::ScintillaCocoa* back = [_sciEditor backend];
  back->RegisterNotifyCallback(nil, notification);

  [_sciEditor setOwner: self];

  if( _fType < 1 ) return; // kick out early uninitialize or error

  switch( _fType )
  {
    case CFILE:
    case CPPFILE:
    case OBJCFILE:
    case OBJCPPFILE:
    case HDFILE:
    case JFILE:
    {
      ed = CScintillaController::getInstance();
      break;
    }
    case TXTFILE:
    {
      ed = TXTScintillaController::getInstance();
      break;
    }
    case MKFILE:
    {
      ed = MKScintillaController::getInstance();
      break;
    }
    case SQLFILE:
    {
      ed = SQLScintillaController::getInstance();
      fflush(stdout);
      break;
    }
    default:
    ;
  }
  if( ed )
  {
    ed->setDefaultEditor(_sciEditor,_fType);
  }
  
  [_findField display];
}

/* we use the MyEditor.nib file for our main window */
- (NSString *)windowNibName
{
    return @"MyEditor";
}

/* after the window has been setup, put the text on if there is any
 * set a save point scintilla can tell us if the document is dirty
 */
- (void)windowControllerDidLoadNib:(NSWindowController *) aController
{
  [super windowControllerDidLoadNib:aController];

  if( _fileData )
  {
    NSString* text = [[NSString alloc]initWithData:_fileData
                                          encoding: NSUTF8StringEncoding];
    [_sciEditor setString:text];
    [text release];
  }

  [_sciEditor setGeneralProperty: SCI_EMPTYUNDOBUFFER parameter: 0 value: 0];
}

/* method used when we want to save */
- (NSData *)dataOfType:(NSString *)typeName error:(NSError **)outError
{
  [_sciEditor setGeneralProperty: SCI_SETSAVEPOINT parameter: 0 value: 0];
  
  NSString* text = [_sciEditor string];
  return [text dataUsingEncoding:NSUTF8StringEncoding];
  return nil;
}

/* method used when we load a file */
- (BOOL)readFromData: (NSData *)data
              ofType:(NSString *)typeName
               error:(NSError **)outError
{
    const char* fType = [typeName UTF8String];

    if( !strcmp(fType,"CFile") )
    {
      _fType = CFILE;
    }
    else if( !strcmp(fType,"CPPFile") )
    {
      _fType = CPPFILE;
    }
    else if( !strcmp(fType,"ObjCFile") )
    {
      _fType = OBJCFILE;
    }
    else if( !strcmp(fType,"ObjCppFile") || !strcmp(fType,"XCFile") )
    {
      _fType = OBJCPPFILE;
    }
    else if( !strcmp(fType,"HFile") )
    {
      _fType = HDFILE;
    }
    else if( !strcmp(fType,"JFile") )
    {
      _fType = JFILE;
    }
    else if( !strcmp(fType,"TXTFile") )
    {
      _fType = TXTFILE;
    }
    else if( !strcmp(fType,"MKFile") )
    {
      _fType = MKFILE;
    }
    else if( !strcmp(fType,"SQLFile") )
    {
      _fType = SQLFILE;
    }
    else
    {
      _fType = FILE_TYPE_ERROR;
    }

    _fileData = [data retain];
    return YES;
}

-(IBAction)hideFind:(id)sender
{
  [_findField setHidden:true];
  [_lineField setHidden:true];
  [_myWindow makeFirstResponder:[_sciEditor content]];
}

-(IBAction)findForward:(id)sender
{
  NSString* str = [_findField stringValue];

  ScintillaController* con = ScintillaController::getInstance();
  con->findText(_sciEditor,[str UTF8String],DOWN,false);
}

-(IBAction)findBackward:(id)sender
{
  NSString* str = [_findField stringValue];

  ScintillaController* con = ScintillaController::getInstance();
  con->findText(_sciEditor,[str UTF8String],UP,false);
}

-(IBAction)findText:(id)sender
{
  [self findForward:sender];
}

-(IBAction)goToLine:(id)sender
{
  int line = [_lineField intValue];

  ScintillaController* con = ScintillaController::getInstance();

  con->goToLine(_sciEditor,line-1);

  [_myWindow makeFirstResponder:[_sciEditor content]];

  [_lineField setHidden:true];
}

-(IBAction)showFind:(id)sender
{
  [_findField setHidden:false];
  [_lineField setHidden:true];
  [_myWindow makeFirstResponder:_findField];
}

-(IBAction)showGoto:(id)sender
{
  [_lineField setHidden:false];
  [_findField setHidden:true];
  [_myWindow makeFirstResponder:_lineField];
}

-(IBAction)wrap:(id)sender
{
  ScintillaSend* send = ScintillaSend::getInstance();
  if( _isWrap )
  {
    _isWrap = false;
    send->set(_sciEditor,SCI_SETWRAPMODE,SC_WRAP_NONE,0);
  } else {
    _isWrap = true;
    send->set(_sciEditor,SCI_SETWRAPMODE,SC_WRAP_WORD,0);
  }
}

-(IBAction)foldAll:(id)sender
{
  ScintillaController* cont = ScintillaController::getInstance();

  cont->foldAll(_sciEditor);
}

-(IBAction)unfoldAll:(id)sender
{
  ScintillaController* cont = ScintillaController::getInstance();

  cont->unfoldAll(_sciEditor);
}  
@end

/*
Copyright (c) 2010, Michael Mullin <masmullin@gmail.com>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by Michael Mullin <masmullin@gmail.com>.
4. Neither the name of Michael Mullin nor the
   names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY MICHAEL MULLIN ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL MICHAEL MULLIN BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
